"""An empty file where tests should be
"""

import pytest
import distob

#TODO Use pytest fixtures to reliably create an IPython local cluster for tests
#     (check how the IPython test suite does it)

def test_engines():
    pass
